import matplotlib.pyplot as plt
import numpy as np
from scipy.io.wavfile import read as wav_read
from scipy.io.wavfile import write as wav_write
from scipy.signal import fftconvolve,decimate,dlti,lfilter,cheby1
from scipy import interpolate

from matplotlib.ticker import MaxNLocator
from matplotlib import rc


from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm
from matplotlib import gridspec
import matplotlib.ticker as tck
from matplotlib.legend_handler import HandlerPatch
import matplotlib.patches as mpatches

def make_legend_arrow(legend, orig_handle,
                      xdescent, ydescent,
                      width, height, fontsize):
    p = mpatches.FancyArrow(0, 0.5*height, width, 0, length_includes_head=True, head_width=0.75*height )
    return p


rc('text', usetex=False)
#rc('font',**{'family':'sans-serif','sans-serif':['Helvetica']})

def plot_continous(n,xn,label='',show_yaxis=False,ylim=None,title=None,color='black'):
    """Grafica de una secuencia de numeros reales y la reconstruccion ideal
    de la senal en tiempo continuo.
    
    Parameters
    ----------
    
    n: numpy.array [shape=(signal_length, )]    
        Vector con los valores del tiempo discreto
    
    xn: numpy.array [shape=(signal_length, )]    
        Vector con los valores de la secuencia x[n]
    
    label: string
        Etiqueta de la grafica
        (Default value=None)
       
    show_yaxis: bool
        Si es True muestra los valores del eje y
        (Default value=False) 
 
    ylim: list [y_min, y_max]
        Fija los limites del eje y de la grafica
        (Default value=None)  
        
    title: string
        Titulo de la grafica
        (Default value=None)
               
    
    Returns
    -------
    No retorna nada
    
    """

    plt.plot(n,xn,'-',color=color,label=label)
    plt.xlabel('tiempo discreto (n)') #,fontsize=18
    #plt.ylabel('x[n]',fontsize=18)
    if label != '':
        plt.legend(prop={'size': 10})
    ax = plt.gca()
    if show_yaxis==False:
        ax.yaxis.set_visible(False)
    if ylim!=None:
        ylim = ax.set_ylim(ylim)
    ax.margins(0.05)
    if title != None:
        ax.set_title(title)
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))


def plot_deltas(n,xn,label='',show_yaxis=False,ylim=None,title=None,color='red'):
    """Grafica tren de deltas en posiciones n con alturas xn.
    
    Parameters
    ----------
    
    n: numpy.array [shape=(signal_length, )]    
        Vector con los valores del tiempo discreto
    
    xn: numpy.array [shape=(signal_length, )]    
        Vector con los valores de la secuencia x[n]
    
    label: string
        Etiqueta de la grafica
        (Default value=None)
       
    show_yaxis: bool
        Si es True muestra los valores del eje y
        (Default value=False) 
 
    ylim: list [y_min, y_max]
        Fija los limites del eje y de la grafica
        (Default value=None)  
        
    title: string
        Titulo de la grafica
        (Default value=None)
               
    
    Returns
    -------
    No retorna nada
    
    """
    ax = plt.gca()
    for j in range(len(n)):
        arrow = ax.arrow(n[j],0,0,xn[j],head_width=0.3,head_length=0.05,width=0.1,head_starts_at_zero=False,length_includes_head=True)
#    plt.step(n,xn)
    plt.xlabel('tiempo continuo (s)') #,fontsize=18
    #plt.ylabel('x[n]',fontsize=18)
    if label != '':
        plt.legend(prop={'size': 10})
    plt.legend([arrow,], [label,], handler_map={mpatches.FancyArrow : HandlerPatch(patch_func=make_legend_arrow),})
 
    if show_yaxis==False:
        ax.yaxis.set_visible(False)
    if ylim!=None:
        ylim = ax.set_ylim(ylim)
    ax.margins(0.05)
    if title != None:
        ax.set_title(title)
    #ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    ax.xaxis.set_major_formatter(tck.FormatStrFormatter('%g $T$'))
    ax.xaxis.set_major_locator(tck.MultipleLocator(base=3.0))

def plot_seqs(n,xn,label=None,plot_lines=False,show_yaxis=False,ylim=None,title=None,force_int_ticks=False):
    """Grafica de una secuencia de numeros reales
    
    Parameters
    ----------
    
    n: numpy.array [shape=(signal_length, )]    
        Vector con los valores del tiempo discreto
    
    xn: numpy.array [shape=(signal_length, )]    
        Vector con los valores de la secuencia x[n]
    
    label: string
        Etiqueta de la grafica
        (Default value=None)
       
    show_yaxis: bool
        Si es True muestra los valores del eje y
        (Default value=False) 
 
    ylim: list [y_min, y_max]
        Fija los limites del eje y de la grafica
        (Default value=None)  
        
    title: string
        Titulo de la grafica
        (Default value=None)
       
    force_int_ticks: bool
        Si es True se fuerza que los ticks del eje y sean en valores enteros.
        (Default value=False)
        
    
    Returns
    -------
    No retorna nada
    
    """

    if n is not None:
        if label == None:
            plt.stem(n,xn,color='black',linefmt='k-', markerfmt='ko',basefmt='k-')
        else:
            plt.stem(n,xn,label=label,color='black',linefmt='k-', markerfmt='ko',basefmt='k-')
            plt.legend(prop={'size': 10})
        if plot_lines == True:
            plt.plot(n,xn,'-',color='red',label='$x_a(t)$')
    else: 
        if label == None:
            plt.stem(xn,color='black',linefmt='k-', markerfmt='ko',basefmt='k-')
        else:
            plt.stem(xn,label=label,color='black',linefmt='k-', markerfmt='ko',basefmt='k-')
            plt.legend(prop={'size': 10})
        if plot_lines == True:
            plt.plot(xn,'-',color='red',label='$x_a(t)$')
    plt.xlabel('tiempo discreto (n)') #,fontsize=18
	#plt.ylabel('x[n]',fontsize=18)

    ax = plt.gca()
    if show_yaxis==False:
        ax.yaxis.set_visible(False)
    if ylim!=None:
        ylim = ax.set_ylim(ylim)
    ax.margins(0.05)
    if title != None:
        ax.set_title(title)
    if force_int_ticks == True:
        yint = range(int(np.amin(xn)), int(np.ceil(np.amax(xn))+1))
        ax.set_yticks(yint)

    ax.xaxis.set_major_locator(MaxNLocator(integer=True))

def plot_complexs(x,title='', rmax=2,rticks=[0.5, 1, 1.5, 2]):
    """Grafica polar de una secuencia de numeros complejos
    
    Parameters
    ----------
    
    x: numpy.array [shape=(signal_length, )]    
        Vector con los valores de la secuencia x[n]
    
    title: string
        Titulo de la grafica
        (Default value='')
        
    rmax: int > 0
        Maximo modulo de los numeros complejos que se grafica
        (Default value=2)
        
    rticks: list
        Ticks que se generan en la grafica
        (Default value=[0.5, 1, 1.5, 2])
    
    Returns
    -------
    No retorna nada
    
    """

    ax = plt.subplot(111, projection='polar')
    ax.plot(np.angle(x), np.abs(x),'o-',color='black')
    ax.set_title(title)
    ax.set_rmax(rmax)
    ax.set_rticks(rticks)  # less radial ticks
    ax.set_rlabel_position(-22.5)  # get radial labels away from plotted line
    ax.grid(True)
    plt.show()

def gen_signals(type, n_0, n_end, n_c, N=1, A=1, kernel=[1,1,1]):
    """Genera senales de diferentes tipos utiles para el curso. Tambien genera
    un vector de tiempo discreto.
    
    Parameters
    ----------
    
    type: string
        Define el tipo de senal a generar:
        'rect': pulso rectangular que comienza en n_c y tiene ancho N
        'triang': pulso triangular que comienza en n_c y tiene ancho N
        'escalon': escalon que comienza en n_c
        'delta': impulso posicionado en n_c
        'uds': pulso definido por el usuario en kernel. Comienza en n_c.
        'hanning': ventana de hanning que comienza en n_c y tiene ancho N.
        'hamming': ventana de hamming que comienza en n_c y tiene ancho N.
        
    n_0: int
        Valor de comienzo del tiempo discreto.
            
    n_end: int
        Valor de final del tiempo discreto.  
  
    n_c: int
        Tiempo en el que comienza la senal.  
    
    N: int
        Numero de muestras del pulso rectangular y triangular
        (Default value=1)    

    A: int
        Amplitud de la senal generada
        (Default value=1) 

    kernel: list
        Pulso definido por el usuario en caso de utilizar type='uds'.
        (Default value=[1,1,1]) 

    Returns
    -------
    
    n: numpy.array [shape=(n_end-n_0,)]
        Vector de tiempo discreto que comienza en n_0 y termina en n_end
        
    signal: numpy.array [shape=(n_end-n_0,)]
        Senal generada.

    
    """
    n = np.arange(n_0,n_end+1)
    signal = np.zeros_like(n,dtype=float)
    i_c = int(np.argwhere(n==n_c))
    if type == 'rect':
        signal[i_c:i_c+N +1] = 1
    if type == 'triang':
        for j in range(i_c,i_c+N//2+1):
            signal[j] = j - i_c +1
            m = max(signal)
            i = 1
            for j in range(i_c+N//2+1,i_c+N):
                signal[j] = m - i
                i += 1
    if type == 'escalon':
        signal[i_c:] = 1 
    if type == 'delta':
        signal[i_c] = 1 
    if type == 'uds':
        signal[i_c:i_c+len(kernel)] = kernel 
    if type == 'hanning':
        signal[i_c:i_c+N +1] = hanning(N+1)
    if type == 'hamming':
        signal[i_c:i_c+N +1] = hamming(N+1)
        
    signal = A* signal / float(max(signal))
    return n,signal

def retardar(x, ret):
    """Implementa un Sistema que retarda una senal un numero entero de muestras.
    Devuelve x[n - ret]
    
    Parameters
    ----------
    
    x: numpy.array [shape=(signal_length,)]
        Senal de entrada x[n]
    
    ret: int
        Numero de muestras para retardar
    
    Returns
    -------
    
    y: numpy.array [shape=(signal_length,)]
        Vector con la senal de salida y[n] = x[n-ret]

    
    """

    a = np.zeros((abs(ret),))
    if ret > 0:
        y = np.concatenate((a,x),axis=0)
        y = y[:len(x)]
    else:
        y = np.concatenate((x,a),axis=0)
        y = y[abs(ret):]
    return y

def slit(x,h=[0.5, 1, 0.5]):
    """Implementa un Sistema Lineal Invariante en el Tiempo definido por
    su respuesta al impulso h[n]
    
    Parameters
    ----------
    
    x: numpy.array [shape=(signal_length,)]
        Senal de entrada x[n]
    
    h: numpy.array or list [shape=(h_length,)]
        Respuesta al impulso h[n]
        (Default value=[0.5, 1, 0.5])
    
    Returns
    -------
    
    y: numpy.array [shape=(signal_length,)]
        Vector con la senal de salida y[n]

    
    """

    if len(x) < 500:
        y = np.convolve(x,h,mode='same')
    else:
        y = fftconvolve(x,h)
    return y

def wavread(file):
    """Lee un archivo de audio
    
    Parameters
    ----------
    
    file: string
        Nombre del archivo.
    
    
    Returns
    -------
    x: numpy.array [shape=(signal_length,N_chanels )]
        Vector con la senal que se lee en el archivo
        
    fs: int > 0
        Frecuencia de muestreo.
    
    """
            
    
    [fs, x] = wav_read(file)
    x = x/32767.
    return x, fs

def wavwrite(file,data,fs):
    """Guarda un archivo de audio
    
    Parameters
    ----------
    
    file: string
        Nombre del archivo.
        
    data: numpy.array [shape=(signal_length,N_chanels )]
        Vector con la senal que se desea guardar.
        
    fs: int > 0
        Frecuencia de muestreo.

    
    Returns
    -------
    No retorna nada
    
    """
        
    data = 0.9*data/np.amax(data)
    wav_write(file,fs,data)

def plot_H(omega,H):
    """Grafica una respuesta en frecuencia H(e^jw)
    
    Parameters
    ----------
    
    omega: numpy.array [shape=(signal_length, )]    
        Vector con los valores de omega donde se calculo la H(e^jw)
    
    H: numpy.array [shape=(signal_length, )]
        Vector con la respuesta en frecuencia H(e^jw)


    
    Returns
    -------
    No retorna nada
    
    """

    pi = np.pi
    plt.subplot(2,1,1)
    plt.title('$|H(e^{j\omega})|$')
    plt.xlabel('$\omega$ (rad)')
    plt.ylabel('Modulo')
    plt.plot(omega,np.abs(H))
    plt.xticks([-pi, 0, pi],['$-\pi$','$0$','$\pi$'])
    plt.subplot(2,1,2)
    plt.title('$ < H(e^{j\omega})$')
    plt.plot(omega,np.angle(H))
    plt.xticks([-pi, 0, pi],['$-\pi$','$0$','$\pi$'])
    plt.yticks([-pi, 0, pi],['$-\pi$','$0$','$\pi$'])
    plt.xlabel('$\omega$ (rad)')
    plt.ylabel('Fase')
    plt.tight_layout()

def TdF(x):
    """Calcula Transformada de Fourier de x. Para estimar la TdF se utiliza
    la Fast Fourier Transform (fft).
    
    Parameters
    ----------
    x: numpy.array [shape=(signal_length, )]
        Vector con la senal en tiempo discreto x[n]


    
    Returns
    -------
    X: numpy.array [shape=(signal_length, )]
        Vector con la transformada de Fourier X(e^jw)
    
    """

    X = np.fft.fft(x)
    X = np.fft.fftshift(X)
    return(X)

def plot_TdF(X,phase=False,ticks={'pos': None, 'values': None},name='X',loga=False):
    """Grafica el vector X asumiendo que es una TdF calculada con TdF(x)
    
    Parameters
    ----------
    X: numpy.array [shape=(signal_length, )]
        Vector con TdF
    
    phase: bool
        Grafica tambien la fase de la transformada de fourier
        (Default value=False)
    
    ticks: dict [shape=(N_ticks)]
        Si no es None se usan las posiciones de los ticks en el eje x definidos
        en ticks['pos'] y los valores definidos en ticks['values']
        (Default value={'pos': None, 'values': None})
    
    
    Returns
    -------
    No retorna nada.
    
    """

    pi = np.pi
    omega = np.linspace(-pi+2*pi/len(X),pi,len(X))
    if phase == True:
        plt.subplot(2,1,1)
    if loga == False:
        plt.plot(omega,np.abs(X))
    else:
        plt.plot(omega,np.log(np.abs(X)))        
    plt.title('$|'+name+'(e^{j\omega})|$')
    if ticks['pos']==None:
        plt.xticks([-pi, 0, pi],['$-\pi$','$0$','$\pi$'])
    else:
        plt.xticks(ticks['pos'],ticks['values'])			
    plt.xlabel('$\omega$ (rad)')
    plt.ylabel('Modulo')

    if phase == True:
        plt.subplot(2,1,1)    
        plt.subplot(2,1,2)
        plt.plot(omega,np.angle(X))
        plt.title('$ < '+name+'(e^{j\omega})$')
        if ticks['pos']==None:
            plt.xticks([-pi, 0, pi],['$-\pi$','$0$','$\pi$'])
        else:
            plt.xticks(ticks['pos'],ticks['values'])			
        plt.xlabel('$\omega$ (rad)')
        plt.ylabel('Fase')
        plt.tight_layout()

def plot_TdF_fs(X,fs,phase=False,ticks={'pos': None, 'values': None},name='X',loga=False):
    """Grafica el vector X asumiendo que es una TdF calculada con TdF(x) y utilizando vector de frecuencias continuas.
    
    Parameters
    ----------
    X: numpy.array [shape=(signal_length, )]
        Vector con TdF
    
    phase: bool
        Grafica tambien la fase de la transformada de fourier
        (Default value=False)
    
    ticks: dict [shape=(N_ticks)]
        Si no es None se usan las posiciones de los ticks en el eje x definidos
        en ticks['pos'] y los valores definidos en ticks['values']
        (Default value={'pos': None, 'values': None})
    
    
    Returns
    -------
    No retorna nada.
    
    """

    pi = np.pi
    fN = fs/2.
    X = X/float(fs) #Normalizo por frecuencia de muestreo
    freqs = np.linspace(-fN+2*fN/len(X),fN,len(X))
    if phase == True:
        plt.subplot(2,1,1)
    if loga == False:
        plt.plot(freqs,np.abs(X))
    else:
        plt.plot(freqs,np.log(np.abs(X)))
    plt.title('$|'+name+'(f)|$')
#    if ticks['pos']==None:
#        plt.xticks([-fN, 0, fN],['$-\fs/2 = %d$' % -fN,'$0$','$\fs/2 = %d$' % fN])
#    else:
#        plt.xticks(ticks['pos'],ticks['values'])			
    plt.xlabel('frecuencia (Hz)')
    plt.ylabel('Modulo')

    if phase == True:
        plt.subplot(2,1,1)    
        plt.subplot(2,1,2)
        plt.plot(omega,np.angle(X))
        plt.title('$ < '+name+'(f)$')
        if ticks['pos']==None:
            plt.xticks([-fN, 0, fN],['$-\fs/2 = %d$' % -fN,'$0$','$\fs/2 = %d$' % fN])
        else:
            plt.xticks(ticks['pos'],ticks['values'])			
        plt.xlabel('frecuencia (Hz)')
        plt.ylabel('Fase')
        plt.tight_layout()
        
def plot_Z_3d(x,y,X):
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.plot_surface(x, y, abs(X), rstride=1, cstride=1, cmap=cm.jet)
    ax.set_xlabel('$Re(z)$')
    ax.set_ylabel('$Im(z)$')
    ax.set_zlabel('$|X(z)|$')
    return ax

def plot_Z_3d_ejw(x,y,X,polos=[],ceros=[]):
    pi = np.pi
    fig = plt.figure()
    gs = gridspec.GridSpec(2, 1, height_ratios=[3, 1])
    
    #ax = fig.add_subplot(211, projection='3d')
    ax = plt.subplot(gs[0], projection='3d')
    ax.plot(x, y, abs(X),'r')

    omega = np.linspace(-pi,pi,200)
    z = np.exp(1j*omega) # Evalua la Transformada Z solo en e^{jw}

    ax.plot(np.real(z),np.imag(z),0,'k')
    for p in polos:
        ax.text(np.real(p),np.imag(p),0,'x',zdir='y')
    for c in ceros:
        ax.text(np.real(c),np.imag(c),0,'o',zdir='y')   
    #ax.scatter(1,0)
    
    ax.set_xlabel('$Re(z)$')
    ax.set_ylabel('$Im(z)$')
    ax.set_zlabel('$|X(z)|$')
    
    #ax2 = fig.add_subplot(212)
    ax2 = plt.subplot(gs[1])
    omega = np.linspace(-pi+2*pi/len(X),pi,len(X))
    ax2.plot(omega,np.abs(X),'r')
    ax2.set_title('$|X(e^{j\omega})|$')
    ax2.set_xticks([-np.pi, 0, np.pi],['$-\pi$','$0$','$\pi$'])			
    ax2.set_xlabel('$\omega$ (rad)')
    ax2.set_ylabel('Modulo')
    
    return [ax,ax2]

def plot_Z_2d(x,y,X):
    plt.figure()
    plt.imshow(abs(X), extent=[np.amin(x),np.amax(x),np.amin(y),np.amax(y)], cmap=cm.jet,origin='lower')
    
def muestrear(t,xc,Omega_s):
    pi = np.pi
    fs = Omega_s/2./pi
    Ts = 1/float(fs)
    T_vector = t[1]-t[0]
    f = interpolate.interp1d(t, xc, kind='cubic')
    t_new = np.arange(t[0],t[-1],Ts)
    x = f(t_new)
    n_min = np.floor(t_new[0]/Ts)
    n_max = np.ceil(t_new[-1]/Ts)
    n = np.arange(n_min,n_max)
    if n.shape < x.shape:
        n = np.arange(n_min,n_max+1)
    return n,x

def reconstruir(t,x,Omega_s):
    pi = np.pi
    fs = Omega_s/2./pi
    Ts = 1/float(fs)
    t_muestras = np.arange(t[0],t[-1],Ts)
    f = interpolate.interp1d(t_muestras, x, fill_value='extrapolate')
    xtemp = f(t)
    xs = np.zeros_like(xtemp)
    for j,t1 in enumerate(t_muestras):
        i = np.argmin(np.abs(t-t1))
        xs[i] = x[j]
    hr = np.sinc(t/Ts)
    xr = np.convolve(xs,hr,mode='same')
    return xr

def decimador(x,M,fs):
    filter = dlti([1],[1])
    xd = decimate(x,M,ftype=filter)
    return xd,fs/float(M)

def expansor(x,L,fs):
    xe = np.zeros((L*len(x),))
    for j in range(0,len(x)):
        xe[int(L*j)] = x[j]
    return xe,fs*L

def pasabajos(x,wc,G=1):
    pi = np.pi
    W = 0.8 * wc/pi
    n = 8
    system = dlti(*cheby1(n, 0.05, W))
    b, a = system.num, system.den
    b = b*G
    xf = lfilter(b,a,x)
    return(xf)

def autocorrelacion(x):
    c = np.correlate(x, x,'same')/float(len(x))
    return c
